#include "hit_map.h"
#include "util.h"

/**
 * Creates a map.
 * @param size
 * @return The address of the created map
 */
HitMapType* hit_map__create(int size, int indel_count) {
  /* for each read, memorize MAP_DETAIL_SIZE best scores */
  HitMapType* map;
  SAFE_FAILURE__ALLOC(map, 1, HitMapType);
  hit_map__init(map, size, indel_count);
  return map;
}

/**
 * Init a map structure.
 * @param map The map structure (already allocated)
 * @param size The map size (at the total number of reads)
 * @param indel_count The maximal number of indels (when aligning a read)
 */
void hit_map__init(HitMapType* map, int size, int indel_count) {
  map->size = size;
  map->indel_count = indel_count;
  map->mapped = 0;
  /* for each read, memorize MAP_DETAIL_SIZE best scores */
  SAFE_FAILURE__ALLOC(map->map, size, HitInfoType*);
  int read_id;
  for (read_id = 0; read_id < size; ++read_id) {
    SAFE_FAILURE__ALLOC(map->map[read_id], MAP_DETAIL_SIZE, HitInfoType);
    int k;
    for(k = 0; k < MAP_DETAIL_SIZE; ++k) {
      map->map[read_id][k].score             = MIN_SCORE;
      map->map[read_id][k].adjusted_score    = MIN_SCORE;
      map->map[read_id][k].ref_pos           = UNMAPPED;
      map->map[read_id][k].ref_id            = UNMAPPED;
      map->map[read_id][k].sense             = 0;
      map->map[read_id][k].traceback         = NULL;
      map->map[read_id][k].traceback_seq_len = 0;
      map->map[read_id][k].indel_positions   = NULL;
    }
  }
}

/**
 * Updates a map with a new alignment
 * @param map The map to update
 * @param read_id The aligned read's index in the reads database
 * @param ref_pos The position where the aligned part of the reference starts
 * @param alignment The obtained alignment
 * @return The rank of the alignment if its score was good enough to include in the map
 * (between 0 and MAP_DETAIL_SIZE exclusive), or UNMAPPED (-1) otherwise.
 */
int hit_map__update(HitMapType* map, const int read_id, const int ref_id, const int ref_pos, const AlignmentType* alignment, int alignment_sense) {
  HitInfoType* info = map->map[read_id];
  ScoreType score = alignment__get_score(alignment, alignment->best_i, alignment->best_j);
  /* if the update is successful, and the read wasn't mapped before, to the number of mapped reads */
  int mapped = (info[0].score == MIN_SCORE) ? 1 : 0;

  int i = MAP_DETAIL_SIZE - 1;
  while (i >= 0 && score > info[i].score) i--;

  /* score is smaller, so no update needed */
  if (i ==  MAP_DETAIL_SIZE - 1) {
    return UNMAPPED;
  }

  /* free the last one (allocated structures only) because pushed away ... */
  if (info[MAP_DETAIL_SIZE-1].indel_positions) {
    free(info[MAP_DETAIL_SIZE-1].indel_positions);
    info[MAP_DETAIL_SIZE-1].indel_positions = NULL;
  }
  if (info[MAP_DETAIL_SIZE-1].traceback) {
    free(info[MAP_DETAIL_SIZE-1].traceback);
    info[MAP_DETAIL_SIZE-1].traceback = NULL;
  }

  /* shift to create one insert space */
  int j;
  for (j = MAP_DETAIL_SIZE-1; j > i+1; --j) {    
    info[j] = info[j-1];
  }
  
  /* insert the new element */
  info[j].score   = score;
  info[j].ref_id  = ref_id;
  info[j].ref_pos = (ref_pos + alignment->best_j0);
  info[j].sense   = (alignment_sense & ALIGNMENT_SENSE_REVERSE ? (-1) : 1);
  SAFE_FAILURE__ALLOC(info[j].indel_positions, map->indel_count, short);
  memcpy(info[j].indel_positions, alignment->read_indel_positions, map->indel_count*sizeof(short));
  SAFE_FAILURE__ALLOC(info[j].traceback, TRACEBACK_SEQ_SIZE(alignment->traceback_seq_len), unsigned char);
  info[j].traceback_seq_len = alignment->traceback_seq_len;
  /* copy but reverse the traceback of the alignment ("right order" now) */
  int p;
  for (p = 0; p < alignment->traceback_seq_len; ++p) {
    TRACEBACK_SEQ_SET(info[j].traceback, p, TRACEBACK_SEQ_GET(alignment->traceback, (alignment->traceback_seq_len - 1 - p )));
  }
  map->mapped += mapped;
  return j;
}

/**
 * Obtain the score for a read, when mapped on a given position
 * @param map
 * @param read_id The read index
 * @param ref_id The reference index
 * @param ref_pos The position of interest in the reference genome
 * @param rank The rank of that score will be placed at this address
 * @return The score
 */
inline int hit_map__score_for(const HitMapType* map, const int read_id, const int ref_id, const int ref_pos) {
  int i;
  for (i = 0; i < MAP_DETAIL_SIZE; ++i) {
    if ( map->map[read_id][i].ref_id == ref_id  &&  map->map[read_id][i].ref_pos == ref_pos) {
      return map->map[read_id][i].score;
    }
  }
  return MIN_SCORE;
}

/**
 * Quick display of the successfully matched reads: index, position on the reference, score, adjusted_score [FIXME] and traceback
 * @param map
 */
void hit_map__display(const HitMapType* map) {
  int i;
  for (i = 0; i < map->size; ++i) {
    int j;
    for (j = 0; j < MAP_DETAIL_SIZE && map->map[i][j].ref_pos != UNMAPPED; ++j) {
        printf("read_id: %d  ref_(id,pos): (%d,%d)  score: %d  adj_score: %d  len: %d\n", i, map->map[i][j].ref_id, map->map[i][j].ref_pos, map->map[i][j].score, map->map[i][j].adjusted_score, map->map[i][j].traceback_seq_len);
        TRACEBACK_DISPLAY(map->map[i][j].traceback, map->map[i][j].traceback_seq_len, stdout);
        printf("\n");
    }
  }
}

/**
 * Cleanup
 * @param map
 */
void hit_map__destroy(HitMapType* map) {
  int i;
  for (i = 0; i < map->size; ++i) {
    int j;
    for (j = 0; j < MAP_DETAIL_SIZE; ++j) {
      if (map->map[i][j].indel_positions) {
	free(map->map[i][j].indel_positions);
	map->map[i][j].indel_positions = NULL;
      }
      if (map->map[i][j].traceback) {
        free(map->map[i][j].traceback);
	map->map[i][j].traceback = NULL;
      }
    }
    free(map->map[i]);
    map->map[i] = NULL;
  }
  free(map->map);
  map->map = NULL;
}

