
#include <unistd.h>
#include "reads_data.h"

/**
 * Creates an empty read structure
 */

ReadDataType* read__create(int read_len) {
  ReadDataType* read;
  SAFE_FAILURE__ALLOC(read,1,ReadDataType);
  SAFE_FAILURE__ALLOC(read->info,MAX_HEADER_LEN,char);
  SAFE_FAILURE__ALLOC(read->sequence,COMPRESSED_LEN_N_BYTES(read_len,N_BYTES),CODE_TYPE);
  SAFE_FAILURE__ALLOC(read->quality,COMPRESSED_Q_LEN(read_len),QUAL_TYPE);
  return read;
}

/**
 * Reverse-complement a read
 */
inline void read__reverse(const ReadDataType* src, ReadDataType* dest, int len) {
  /* WARNING: only sequence and quality fields are cloned;
   * also, the dest is considered to have all the allocated space needed
   * by default, the read is compressed
   */
#ifndef NUCLEOTIDES
  dest->first_base = BASE_CODE_COMPLEMENTARY(base_compressed_color_transformation(src->first_base, src->sequence, len, 0));
#endif
  sequence__reverse_compressed(src->sequence, dest->sequence, len);
  if (src->quality) {
    if (!dest->quality) {
      SAFE_FAILURE__ALLOC(dest->quality, COMPRESSED_Q_LEN(len), QUAL_TYPE);
    }
    
#ifndef NUCLEOTIDES
    quality__reverse_compressed(src->first_qual, src->quality, dest->quality, len);
    dest->first_qual = NTH_QUAL(src->quality, len-1);
#else
    quality__reverse_compressed(src->quality, dest->quality, len);
#endif
  } else {
    if (dest->quality) {
      free (dest->quality);
    }
    dest->quality = NULL;
  }
}


/**
 * Destroys a read database
 */
void read__destroy(ReadDataType* read) {
  free(read->info);
  read->info = NULL;
  free(read->sequence);
  read->sequence = NULL;
  if (read->quality) {
    free(read->quality);
    read->quality = NULL;
  }
}




/**
 * Load csfasta reads file
 *
 */

int load_reads_db_fasta_csfasta(const char* reads_filename, const char* quals_filename, ReadsDBType* db) {
  char line[MAX_LINE_LEN];
  char* read_result;
  FILE* reads_in = NULL;
  FILE* quals_in = NULL;
  int   line_len;
  short* quality;

  if (reads_filename)
    reads_in = fopen(reads_filename, "r");

  if (quals_filename)
    quals_in = fopen(quals_filename, "r");

  db->size = 0;
  db->read_len = 0;
  db->name = NULL;

  if (!reads_in) {
    ERROR__(("Failed to read input file %s (read sequences).", reads_filename));
    exit(RETURN_INPUT_ERR);
  }

  /* First scan: count how many reads are there in the file */

  /* Skip comments at the beginning; after this, i will point to the start of the first non-comment line */
  int i_start = 0;
  do {
    i_start = ftell(reads_in);
    read_result = fgets(line, MAX_LINE_LEN, reads_in);
    /* Is there a title in the comment ? */
    if (read_result && strncmp(line, TITLE_LINE_MARKER, TITLE_LINE_MARKER_LEN) == 0) {
      int name_len = strcspn(line + TITLE_LINE_MARKER_LEN, "\r\n");
      SAFE_FAILURE__ALLOC(db->name, (name_len + 1), char);
      strncpy(db->name, line + TITLE_LINE_MARKER_LEN, name_len);
    }
  } while (read_result && get_input_line_type(line) == LINE_TYPE_COMMENT);

  if (!read_result) {
    ERROR__(("Failed to read input file %s (read sequences) first line.", reads_filename));
    exit(RETURN_INPUT_ERR);
  }
  
  /* The last line is not a comment, it is the header of the first sequence */
  do {
    if (get_input_line_type(line) == LINE_TYPE_SEQ_HEADER &&
        fgets(line, MAX_LINE_LEN, reads_in) &&
        get_input_line_type(line) == LINE_TYPE_SEQ) {
      if (db->read_len == 0) {
#ifdef NUCLEOTIDES
        db->read_len = strcspn(line, "\r\n");
#else
        db->read_len = strcspn(line, "\r\n") - 2;
        /* without the \n and the first 2 chars, which are is a base and the first color, used to obtain the real first base */
#endif
        if (db->read_len <= 0) {
          ERROR__(("\nReads file seams corrupt: the first read's length is %d. Aborting.\n\n", db->read_len));
          exit(RETURN_INPUT_ERR);
        }
      } else {
#ifdef NUCLEOTIDES
        if (strcspn(line, "\r\n") != db->read_len) 
#else
        if (strcspn(line, "\r\n") - 2 != db->read_len) 
#endif
          {
            ERROR__(("\nThe read #%d does not have the expected length (length is %d, expected %d).\n"
                     "All the reads must be of equal length for the parallel processing.\n\n",
                     db->size + 1, 
#ifdef NUCLEOTIDES
                     (int)strcspn(line, "\r\n"),
#else
                     (int)strcspn(line, "\r\n") - 2,
#endif               
                     db->read_len));
            exit(RETURN_INPUT_ERR);
          }
      }
      ++db->size;
    }
  } while (fgets(line, MAX_LINE_LEN, reads_in));

  /* The last line is a read; we can obtain the read length */

  /* Allocate */
  SAFE_FAILURE__ALLOC(db->reads, db->size, ReadDataType);

  /* Second scan: load reads */

  /* Go back to the first non-comment line */
  fseek(reads_in, i_start, SEEK_SET);
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loading reads..."););
  {
    int i;
    for (i = 0; i < db->size; ) {
      if (!fgets(line, MAX_LINE_LEN, reads_in)) {
        return RETURN_INPUT_ERR;
      }
      line_len = strlen(line);
      switch (get_input_line_type(line)) {
      case LINE_TYPE_SEQ_HEADER:
	// we suppose that the header is read all at once
	delete_trailing_nl(line);
	/* info size: line_len - 1 (without the >); allocate line_len (1 for \0) */
	SAFE_FAILURE__ALLOC(db->reads[i].info, line_len, char);
	strncpy(db->reads[i].info, line + 1 /* without '>' */, line_len - 1);
	db->reads[i].quality = NULL;
	break;
      case LINE_TYPE_SEQ:
	delete_trailing_nl(line);
#ifdef NUCLEOTIDES
	db->reads[i].sequence = string_to_compressed_code(line);
#else
	// also, we suppose that the sequence is read all at once, since it is known to be short
	// the first base in the file is always the same; the real first base is obtained by transforming it with the first color
	db->reads[i].first_base = BASE_SYMBOL_TO_CODE(line[0]);
	db->reads[i].first_base = TRANSFORM(db->reads[i].first_base, STRIP(line[1]));
	// copy the color sequence after the fake first base and the first color
	db->reads[i].sequence = string_to_compressed_code(line+2);
#endif
	i++;
	break;
      default:
	// ignore other kind of lines
	break;
      }
      VERB_FILTER(VERBOSITY_MODERATE, display_progress(i, db->size, 0))
	}
    fclose(reads_in);
  }
  
  
  /* Try to load qualities */
  if (!quals_in) {
    WARNING__(("Failed to read input file %s (quality of read sequences). All sequences will be considered at maximum quality.\n", quals_filename));
  } else {
    int i;
    VERB_FILTER(VERBOSITY_MODERATE, printf("Loading qualities..."););
#ifndef NUCLEOTIDES
    SAFE_FAILURE__ALLOC(quality, db->read_len + 1, short);
#else
    SAFE_FAILURE__ALLOC(quality, db->read_len, short);
#endif
    /* Load qualities */
    /* For the moment, no rigorous  verification is done, qualities are considered in the right order */    
    for (i = 0; i < db->size; ) {
      if(!fgets(line, MAX_LINE_LEN, quals_in)) {
        break;
      }
      line_len = strlen(line);
      if (get_input_line_type(line) == LINE_TYPE_SEQ) {
#ifndef NUCLEOTIDES
        parse_integer_sequence_to(line, quality, db->read_len + 1);
        db->reads[i].first_qual = QUALITY_LEVEL(quality[0]);
        db->reads[i].quality = compress_quality_sequence(quality, db->read_len, 1);
#else
	parse_integer_sequence_to(line, quality, db->read_len);
        db->reads[i].quality = compress_quality_sequence(quality, db->read_len, 0);
#endif
	i++;
      }      
      VERB_FILTER(VERBOSITY_MODERATE, display_progress(i, db->size, 0));
    }
    fclose(quals_in);    
  }
  return db->size;
}


/**
 * Load fastq reads file
 *
 */
int load_reads_db_fastq(const char* reads_filename, ReadsDBType* db) {
  char line[MAX_LINE_LEN];
  char* read_result;
  FILE* reads_in = fopen(reads_filename, "r");
  int line_len, i_start=0, lineNo=-1, offset=-1, lineType, nextLineType=-1;
  short* quality;

  db->size = 0;
  db->read_len = 0;
  db->name = NULL;

  if (!reads_in) {
    ERROR__(("Failed to read input file %s (read sequences).", reads_filename));
    exit(RETURN_INPUT_ERR);
  }

  /* First scan: count how many reads are there in the file */

  /* Skip comments at the beginning; after this, i will point to the start of the first non-comment line */
  do {
    offset++;
    lineNo++;
    i_start = ftell(reads_in);
    read_result = fgets(line, MAX_LINE_LEN, reads_in);
    /* Is there a title in the comment ? */
    if (read_result && strncmp(line, TITLE_LINE_MARKER, TITLE_LINE_MARKER_LEN) == 0) {
      int name_len = strcspn(line + TITLE_LINE_MARKER_LEN, "\r\n");
      SAFE_FAILURE__ALLOC(db->name, (name_len + 1), char);
      strncpy(db->name, line + TITLE_LINE_MARKER_LEN, name_len);
    }
  } while (read_result && get_input_line_type_fastq(line, lineNo, offset) == LINE_TYPE_COMMENT);

  if (!read_result) {
    ERROR__(("Failed to read input file %s (read sequences) first line.", reads_filename));
    exit(RETURN_INPUT_ERR);
  }

  /* The last line is not a comment, it is the header of the first sequence */
  do {
    if ((lineType = get_input_line_type_fastq(line, lineNo++, offset)) == LINE_TYPE_SEQ_HEADER &&
        fgets(line, MAX_LINE_LEN, reads_in) &&
        (nextLineType = get_input_line_type_fastq(line, lineNo++, offset)) == LINE_TYPE_SEQ) {
      if (db->read_len == 0) {
#ifdef NUCLEOTIDES
        db->read_len = strcspn(line, "\r\n");
#else
        db->read_len = strcspn(line, "\r\n") - 2;
        /* without the \n and the first 2 chars, which are is a base and the first color, used to obtain the real first base */
#endif
        if (db->read_len <= 0) {
          ERROR__(("\nReads file seams corrupt: the first read's length is %d. Aborting.\n\n", db->read_len));
          exit(RETURN_INPUT_ERR);
        }
      } else {
#ifdef NUCLEOTIDES
        if (strcspn(line, "\r\n") != db->read_len)
#else
        if (strcspn(line, "\r\n") - 2 != db->read_len)
#endif
          {
            ERROR__(("\nThe read #%d does not have the expected length (length is %d, expected %d).\n"
                     "All the reads must be of equal length for the parallel processing.\n\n",
                     db->size + 1, 
#ifdef NUCLEOTIDES
                     (int)strcspn(line, "\r\n"),
#else
                     (int)strcspn(line, "\r\n") - 2,
#endif
                     
                     db->read_len));
            exit(RETURN_INPUT_ERR);
          }
      }
      ++db->size;
    } else if ((lineType == LINE_TYPE_SEQ_HEADER && nextLineType != LINE_TYPE_SEQ) || lineType == LINE_TYPE_UNKNOWN) {
      ERROR__(("\nReads file seams corrupt at line %d. Exiting.\n", lineNo));
      exit(RETURN_INPUT_ERR);
    }
  } while (fgets(line, MAX_LINE_LEN, reads_in));
  /* The last line is a read; we can obtain the read length */
  /* Allocate */
  SAFE_FAILURE__ALLOC(db->reads, db->size, ReadDataType);
  SAFE_FAILURE__ALLOC(quality, db->read_len, short);

  /* Second scan: load reads */
  /* Go back to the first non-comment line */
  fseek(reads_in, i_start, SEEK_SET);
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loading reads..."););

  int i = -1;
  lineNo = offset;
  while (i < db->size) {
    if (!fgets(line, MAX_LINE_LEN, reads_in)) {
      if (++i == db->size)
        continue;
      return RETURN_INPUT_ERR;
    }
    line_len = strlen(line);
    switch (get_input_line_type_fastq(line, lineNo++, offset)) {
    case LINE_TYPE_SEQ_HEADER:
      ++i;
      // we suppose that the header is read all at once
      delete_trailing_nl(line);
      /* info size: line_len - 1 (without the >); allocate line_len (1 for \0) */
      SAFE_FAILURE__ALLOC(db->reads[i].info, line_len, char);
      strncpy(db->reads[i].info, line + 1 /* without '@' */, line_len - 1);
      db->reads[i].quality = NULL;
      break;
    case LINE_TYPE_SEQ:
      delete_trailing_nl(line);
#ifdef NUCLEOTIDES
      db->reads[i].sequence = string_to_compressed_code(line);
#else
      // also, we suppose that the sequence is read all at once, since it is known to be short
      // the first base in the file is always the same; the real first base is obtained by transforming it with the first color
      db->reads[i].first_base = BASE_SYMBOL_TO_CODE(line[0]);
      db->reads[i].first_base = TRANSFORM(db->reads[i].first_base, STRIP(line[1]));
      // copy the color sequence after the fake first base and the first color
      db->reads[i].sequence = string_to_compressed_code(line+2);
#endif
      break;
    case LINE_TYPE_QUAL_HEADER:
      break;
    case LINE_TYPE_QUAL:
      // parse the characher string encoding the quality
#ifndef NUCLEOTIDES
      db->reads[i].first_qual = QUALITY_LEVEL(FASTQ_PARSE_QUAL(line[0]));
#endif
      delete_trailing_nl(line);
      memset(quality, 0, (db->read_len)*sizeof(short));
      int p;
#ifdef NUCLEOTIDES
      p = 0;
#else
      p = 1;
#endif
      for (; p < min(db->read_len, strlen(line)); ++p) {
	quality[p] = FASTQ_PARSE_QUAL(line[p]);
      }
#ifdef NUCLEOTIDES
      db->reads[i].quality = compress_quality_sequence(quality, db->read_len, 0);
#else
      db->reads[i].quality = compress_quality_sequence(quality, db->read_len, 1);
#endif
      break;
    default:
      // ignore other kind of lines
      break;
    }
    VERB_FILTER(VERBOSITY_MODERATE, display_progress(i, db->size, 0));
  }
  fclose(reads_in);

  return db->size;
}

int load_reads_db(const char* reads_filename, const char* quals_filename, ReadsDBType* db) {
  if (strcmp(rindex(reads_filename, '.'), ".fastq") == 0) {
    return load_reads_db_fastq(reads_filename, db);
  } else {
    return load_reads_db_fasta_csfasta(reads_filename, quals_filename, db);
  }
}

/**
 * Destroys a read database
 */
void clear_reads_db(ReadsDBType* db) {
  int i;
  for (i = 0; i < db->size; ++i) {
    free(db->reads[i].info);
    free(db->reads[i].sequence);
    if (db->reads[i].quality) {
      free(db->reads[i].quality);
    }
  }
  free(db->reads);
  if (db->name) {
    free(db->name);
  }
}

